//
//  NekoController.mm
//  Neko Type-R

/*
Copyright (c) 2005 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#import "NekoController.h"
#import "NekoView.h"
#import "NekoHelpController.h"
#import "NekoPrefsController.h"


NSString* NekoNotifyNewRaceMessage = @"New race message"; 
extern NSString* NekoTypeRInitialDifficultyKey; 
extern NSString* NekoTypeRBonusFrequencyKey; 
extern NSString* NekoTypeRBonusChanceKey; 

@implementation NekoController

+ (void)initialize
{
	[super initialize]; 
	
	// set up default values for the preferences
	// 
	NSMutableDictionary* defaultValues = [NSMutableDictionary dictionary]; 
	[defaultValues setObject: [NSNumber numberWithInt: kDifficultyNormal] forKey: NekoTypeRInitialDifficultyKey];
	[defaultValues setObject: [NSNumber numberWithFloat: 2.0] forKey: NekoTypeRBonusFrequencyKey];
	[defaultValues setObject: [NSNumber numberWithInt: 4] forKey: NekoTypeRBonusChanceKey];
	
	[[NSUserDefaults standardUserDefaults] registerDefaults: defaultValues]; 
}

- (id)init
{
	if ( self = [super init] )
	{
		helpController = nil; 
		prefsController = nil; 
		
		NSNotificationCenter* nc = [NSNotificationCenter defaultCenter]; 
		[nc addObserver: self selector: @selector(handleNewRaceMessage:) name: NekoNotifyNewRaceMessage object: nil];  
	}
	
	return self; 
}

- (void)dealloc
{
	NSNotificationCenter* nc = [NSNotificationCenter defaultCenter]; 
	[nc removeObserver: self]; 
	
	[helpController release]; 
	[super dealloc]; 
}

- (void)awakeFromNib
{
	[nekoView setValue: self forKey: @"controller"];
}

- (IBAction)newGame: (id)sender
{
	[nekoView startNextRace: YES]; 
}

- (IBAction)displayHelp: (id)sender
{
    if ( !helpController )
    {
		helpController = [[NekoHelpController alloc] init]; 
    }
    
    [helpController showWindow: self];
}

- (IBAction)displayPrefs: (id)sender
{
    if ( !prefsController )
    {
		prefsController = [[NekoPrefsController alloc] init]; 
    }
    
    [prefsController showWindow: self];
}

- (void)updateRaceMessage: (NSString *)message
{
	[raceMessageField setStringValue: message]; 
}

- (void)updatePlayerWordsPerMinute: (UInt32)wpm
{
	[playerWPMField setIntValue: wpm]; 
}

- (void)updateComputerWordsPerMinute: (UInt32)wpm
{
	[computerWPMField setIntValue: wpm]; 
}

- (void)handleNewRaceMessage: (NSNotification *)note
{
	[self updateRaceMessage: [note object]]; 
}

- (BOOL)applicationShouldTerminateAfterLastWindowClosed: (NSApplication *)theApplication 
{
	return YES;
}

@end
