//
//  NekoTyper.mm
//  Neko Type-R

/*
Copyright (c) 2005 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#import "NekoTyper.h"
#import "WordDispenser.h"
#include "randrot-b.h"
#include "Timer.h"
#include "SoundPlayer.h"

#define	kNekoStartingPosition			32.0f
#define kDistanceToRaceStartingPosition	32
#define	kWordLengthMovementMultiplier	4.0f
#define kNekoRunningSpeed				40.0f
#define kNekoSleepAfterXMSeconds		7000
#define kNekoYawnDisplayTime			500
#define kNekoSurpriseDisplayTime		500
#define kNekoRunningImageInterval		200
#define kNekoSleepingImageInterval		1250


extern NSString* NekoNotifyNewRaceMessage;

enum
{
	kNekoActWaiting, 
	kNekoActRunning, 
	kNekoActSleeping, 
	kNekoActSurprised, 
	kNekoActMax
};

@implementation NekoTyper

+ (void)initialize
{
	sdrand(time(NULL)); 
}

- (id)init
{
	if ( self = [super init] )
	{	
		[self reset]; 
		wordToDisplay = [[NSMutableAttributedString alloc] init];
		words = nil; 
		defaultTextAttributes = nil; 
	}
	
	return self; 
}

- (void)dealloc
{
	[wordToType release]; 
	[wordToDisplay release]; 
	[words release]; 
	[defaultTextAttributes release]; 
}

- (NSPoint)position
{
	return position; 
}

- (void)setPosition: (NSPoint)p
{
	position = p; 
}

- (SInt32)imageToDisplay
{
	return imageToDisplay; 
}

- (BOOL)finishedRace
{
	return finishedRace; 
}

- (void)setFinishedRace
{
	imageToDisplay = kNekoImgIdle; 
	currentAction = kNekoActWaiting; 
	finishedRace = YES; 
	[self setValue: @"" forKey: @"wordToType"]; 
	[self updateWordToDisplay: -1]; 
}

- (SInt32)currentPowerup
{
	return currentPowerup; 
}

- (BOOL)shieldedFromBadPowerup
{
	return shieldedFromBadPowerup; 
}

- (BOOL)isMovingIntoStartingPosition
{
	return movingIntoStartingPosition; 
}

- (void)moveNekoForward: (SInt32)move
{
	if ( currentAction == kNekoActSleeping )
	{
		imageToDisplay = kNekoImageSurprised;
		currentAction = kNekoActSurprised; 
		updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoSurpriseDisplayTime; 
		return;
	}
	
	if ( move < 0 )
	{
		if ( movingIntoStartingPosition )
		{
			move = kDistanceToRaceStartingPosition; 
		}
		else
		{
			move = (SInt32)([wordToType length] * kWordLengthMovementMultiplier);
		}
	}
	
	if ( move )
	{
		moveTo = position.x + move; 
		imageToDisplay = kNekoImgRunning1; 
		currentAction = kNekoActRunning; 
		updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoRunningImageInterval; 
	}
}

- (BOOL)isMoving
{
	return currentAction == kNekoActRunning ? YES : NO; 
}

- (void)update
{
	switch(currentAction)
	{
	case kNekoActRunning:
		if ( position.x < moveTo )
		{
			float speedModifier = 1.0f; 
			switch(currentPowerup)
			{
			case kPowerupDistanceBoost:
				speedModifier = 2.5f; 
				break; 
			
			default:
				break;
			}
			
			position.x += gTimer.mGetSecondsSinceLastUpdate() * kNekoRunningSpeed * speedModifier; 
		}
		else
		{
			position.x = (UInt32)moveTo; 
			imageToDisplay = kNekoImgIdle; 
			currentAction = kNekoActWaiting; 
			fallAsleepAtTime = gTimer.mGetMilliseconds() + kNekoSleepAfterXMSeconds;
			currentPowerup = kPowerupMax; 
			
			if ( movingIntoStartingPosition )
			{
				movingIntoStartingPosition = NO; 
			}
			else
			{
				[self setNextWordToType]; 
			}
		}
		
		break; 
	
	case kNekoActWaiting: 
		if ( fallAsleepAtTime < gTimer.mGetMilliseconds() )
		{
			imageToDisplay = kNekoImgYawning; 
			currentAction = kNekoActSleeping; 
			updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoYawnDisplayTime; 
		}
		
		break; 
	
	default:
		break;
	}
	
	if ( updateNextImageAtTime < gTimer.mGetMilliseconds() )
	{
		switch(currentAction)
		{
		case kNekoActRunning: 
			imageToDisplay = imageToDisplay == kNekoImgRunning1 ? kNekoImgRunning2 : kNekoImgRunning1; 
			updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoRunningImageInterval; 
			break;
		
		case kNekoActSleeping:
			imageToDisplay = imageToDisplay == kNekoImageSleeping1 ? kNekoImageSleeping2 : kNekoImageSleeping1; 
			updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoSleepingImageInterval;
			break; 
		
		case kNekoActSurprised:
			[self moveNekoForward: -1]; 
			break; 
		
		default:
			break; 
		}
	}
}

- (void)reset
{
	position = NSMakePoint(kNekoStartingPosition, 0); 
	imageToDisplay = kNekoImageSleeping1; 
	currentAction = kNekoActSleeping;
	moveTo = kNekoStartingPosition; 
	finishedRace = NO; 
	updateNextImageAtTime = 0; 
	fallAsleepAtTime = 0; 
	movingIntoStartingPosition = YES; 
	currentPowerup = kPowerupMax; 
	shieldedFromBadPowerup = NO; 
	
	[self setValue: @"" forKey: @"wordToType"]; 
	[[wordToDisplay mutableString] setString: wordToType];
}

- (void)setNextWordToType
{
	if ( currentPowerup == kPowerupGibberish )
	{
		wordToType = [words getRandomLetters];
	}
	else
	{
		wordToType = [words getWord]; 
	}
	
	[wordToType retain];
	[self updateWordToDisplay: -1]; 
}

- (void)updateWordToDisplay: (SInt32)charactersMatched
{
	if ( charactersMatched < 0 )
	{
		[[wordToDisplay mutableString] setString: wordToType];
		[wordToDisplay setAttributes: defaultTextAttributes range: NSMakeRange(0, [wordToType length])]; 
	}
	else 
	{
		[wordToDisplay removeAttribute: NSForegroundColorAttributeName range: NSMakeRange(0, [wordToDisplay length])]; 
		[wordToDisplay addAttribute: NSForegroundColorAttributeName 
						value: [NSColor whiteColor]
						range: NSMakeRange(0, charactersMatched)];
	}
	
	if ( charactersMatched )
	{
		fallAsleepAtTime = gTimer.mGetMilliseconds() + kNekoSleepAfterXMSeconds;
	}
}

- (void)powerupEffect: (SInt32)type withSound: (BOOL)sound
{
	if ( shieldedFromBadPowerup )
	{
		switch(type)
		{
		case kPowerupForceSleep:
		case kPowerupGibberish:
			if ( sound )
			{
				gSoundPlayer.mPlaySound(kSoundShieldUsed);
			}
			
			shieldedFromBadPowerup = NO; 
			return; 
		
		default:
			break; 
		}
	}
	
	currentPowerup = type; 
	NSString* message = nil; 
	
	switch( type )
	{
	case kPowerupDistanceBoost:
		moveTo += [wordToType length] * kWordLengthMovementMultiplier * 2.0; 
		message = [NSString stringWithFormat: @"%@ just got a speed boost!", name]; 
		break; 
	
	case kPowerupForceSleep:
		moveTo = 0.0f;
		imageToDisplay = kNekoImgYawning; 
		currentAction = kNekoActSleeping; 
		updateNextImageAtTime = gTimer.mGetMilliseconds() + kNekoYawnDisplayTime; 
		[self updateWordToDisplay: 0]; 
		message = [NSString stringWithFormat: @"Whoops. %@ decided to take a nap in the middle of the race.", name]; 
		break; 
	
	case kPowerupGibberish:
		if ( currentAction == kNekoActRunning )
		{
			moveTo = 0.0f; 
			imageToDisplay = kNekoImgIdle; 
			currentAction = kNekoActWaiting; 
			fallAsleepAtTime = gTimer.mGetMilliseconds() + kNekoSleepAfterXMSeconds;
		}
		
		[self setNextWordToType]; 
		message = [NSString stringWithFormat: @"Whoa, it looks like %@ got into some catnip.", name]; 
		break; 
	
	case kPowerupShield:
		shieldedFromBadPowerup = YES; 
		message = [NSString stringWithFormat: @"Haha! %@ is temporarily safe from evil tricks.", name]; 
		break; 
	
	default:
		break; 
	}
	
	if ( message )
	{
		NSNotificationCenter* nc = [NSNotificationCenter defaultCenter]; 
		[nc postNotificationName: NekoNotifyNewRaceMessage object: message]; 
	}
	
	if ( sound )
	{
		switch( type )
		{
		case kPowerupDistanceBoost:
			gSoundPlayer.mPlaySound(kSoundDistanceBoost);
			break; 
		
		case kPowerupForceSleep:
			gSoundPlayer.mPlaySound(kSoundForceSleep);
			break; 
		
		case kPowerupGibberish:
			gSoundPlayer.mPlaySound(kSoundGibberish);
			break; 
		
		case kPowerupShield: 
			gSoundPlayer.mPlaySound(kSoundShield);
			break; 
		
		default:
			break; 
		}
	}
}

@end
