//
//  NekoTypingTeam.mm
//  Neko Type-R

/*
Copyright (c) 2005 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#import "NekoTypingTeam.h"
#import "NekoTyper.h"
#import "NekoView.h"
#include "randrot-b.h"
#include "Timer.h"
#include "SoundPlayer.h"

#define	kFinishLineDistanceFromEdge		40.0f
#define	kBaseUpdateCharactersAtInterval	400
#define	kUpdateCharsDifficultyModifier	25
#define kUpdateWPMTimeInterval			60000
#define kCurrentPowerupDrawPosition		32.0f


extern NSString* NekoNotifyNewRaceMessage;
extern NSString* NekoTypeRInitialDifficultyKey; 
extern NSString* NekoTypeRBonusFrequencyKey; 
extern NSString* NekoTypeRBonusChanceKey;

static inline UInt32 gDistanceToGeneratePowerupAt()
{
	return 30 + (uirand() % 30); 
}

@implementation NekoTypingTeam

+ (void)initialize
{
	sdrand(time(NULL)); 
}

- (id)init
{
	if ( self = [super init] )
	{
		teamID = 0; 
		waitingForRaceToStart = YES; 
		generateNextPowerupAtTime = 0; 
		
		computerControlled = NO; 
		selectedComputerNeko = nil; 
		charactersMatched = 0; 
		updateCharactersMatchedInterval = kBaseUpdateCharactersAtInterval; 
		updateCharactersMatchedAtTime = 0; 
		parentView = nil;
		powerupType = kPowerupMax; 
		
		teamMembers = [[NSMutableArray alloc] init];
		
		NSFont* font = [NSFont fontWithName: @"Monaco" size: 11]; 
		fontHeight = (UInt32)([font boundingRectForFont].size.height + 1); 
		defaultTextAttributes = [NSDictionary dictionaryWithObjectsAndKeys: font,
																			NSFontAttributeName,
																			nil]; 
		[defaultTextAttributes retain]; 
		
		powerupImages[kPowerupDistanceBoost] = [NSImage imageNamed: @"powerup-dist_boost.png"]; 
		powerupImages[kPowerupForceSleep] = [NSImage imageNamed: @"powerup-force_sleep.png"]; 
		powerupImages[kPowerupGibberish] = [NSImage imageNamed: @"powerup-gibberish.png"]; 
		powerupImages[kPowerupShield] = [NSImage imageNamed: @"powerup-shield.png"]; 
		
		numberOfWordsTyped = 0; 
		updateWordsPerMinuteAtTime = 0; 
		wordsPerMinute = [[NSMutableArray alloc] init]; 
	}
	
	return self; 
}

- (void)dealloc
{
	[teamMembers release]; 
	[defaultTextAttributes release]; 
}

- (void)loadTeamImages: (NSString *)nameBase
{
	// load the images for the neko typers
	// 
	nekoImages[kNekoImgIdle] = [NSImage imageNamed:	[NSString stringWithFormat:	@"%@-idle.png", nameBase]]; 
	nekoImages[kNekoImgRunning1] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-run_01.png", nameBase]]; 
	nekoImages[kNekoImgRunning2] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-run_02.png", nameBase]]; 
	nekoImages[kNekoImgYawning] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-yawn.png", nameBase]]; 
	nekoImages[kNekoImageSleeping1] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-sleep_01.png", nameBase]]; 
	nekoImages[kNekoImageSleeping2] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-sleep_02.png", nameBase]]; 
	nekoImages[kNekoImageSurprised] = [NSImage imageNamed: [NSString stringWithFormat:	@"%@-surprise.png", nameBase]]; 
}

- (void)addTeamMember: (NekoTyper *)neko
{
	if ( neko )
	{
		[teamMembers addObject: neko]; 
		[neko setValue: defaultTextAttributes forKey: @"defaultTextAttributes"]; 
	}
}

- (void)updateTeam
{
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		[neko update]; 
	}
	
	if ( waitingForRaceToStart )
	{
		NekoTyper* neko = nil; 
		NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
		while ( neko = (NekoTyper *)[nekoIter nextObject] )
		{
			if ( ![neko isMovingIntoStartingPosition] )
			{
				[neko setNextWordToType]; 
				waitingForRaceToStart = NO; 
			}
			else
			{
				waitingForRaceToStart = YES; 
			}
		}
		
		if ( !waitingForRaceToStart )
		{
			generateNextPowerupAtTime = gTimer.mGetMilliseconds() + powerupGenerationInterval; 
			updateWordsPerMinuteAtTime = gTimer.mGetMilliseconds() + leftoverTimeInMinute; 
			
			if ( computerControlled )
			{
				[self startComputerControl]; 
			}
		}
	}
	else
	{
		if ( computerControlled )
		{
			[self updateComputerControl]; 
		}
		
		if ( generateNextPowerupAtTime < gTimer.mGetMilliseconds() )
		{
			generateNextPowerupAtTime = gTimer.mGetMilliseconds() + powerupGenerationInterval;
			
			if ( powerupType == kPowerupMax && uirand() % chanceOfPowerupAtInterval == 0 )
			{
				nekoWithPowerup = [teamMembers objectAtIndex: uirand() % [teamMembers count]]; 
				powerupPosition = [nekoWithPowerup position].x + gDistanceToGeneratePowerupAt(); 
				powerupType = uirand() % kPowerupMax; 
				if ( powerupPosition > [parentView bounds].size.width - kFinishLineDistanceFromEdge - [powerupImages[powerupType] size].width )
				{
					powerupType = kPowerupMax; 
					generateNextPowerupAtTime = 0; 
				}
			}
		}
		
		if ( updateWordsPerMinuteAtTime && updateWordsPerMinuteAtTime < gTimer.mGetMilliseconds() )
		{
			[wordsPerMinute addObject: [NSNumber numberWithUnsignedLong: numberOfWordsTyped]]; 
			
			UInt32 totalWPM = 0; 
			NSNumber* wpm = nil; 
			NSEnumerator* wpmIter = [wordsPerMinute objectEnumerator]; 
			while ( wpm = (NSNumber *)[wpmIter nextObject] )
			{
				totalWPM += [wpm unsignedLongValue]; 
			}
			
			[parentView updateWordsPerMinute: totalWPM / [wordsPerMinute count] ofTeam: self]; 
			updateWordsPerMinuteAtTime = gTimer.mGetMilliseconds() + kUpdateWPMTimeInterval;
			numberOfWordsTyped = 0; 
		}
	}
}

- (float)drawTeamAtHeight: (float)yAdjustment
{
	float finishLine = [parentView bounds].size.width - kFinishLineDistanceFromEdge; 
	
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		// draw the neko's name
		// 
		NSString* name = [neko valueForKey: @"name"]; 
		[name drawAtPoint: NSMakePoint(2, yAdjustment + 1) withAttributes: defaultTextAttributes]; 
		
		// get the image and word to display
		// 
		NSImage* nekoImage = nekoImages[[neko imageToDisplay]]; 
		NSAttributedString* word = [neko valueForKey: @"wordToDisplay"]; 
		
		// set where the two are going to draw
		// 
		NSPoint drawPos = [neko position]; 
		drawPos.y += yAdjustment; 
		
		// check if this neko typer has reached the end and adjust the image accordingly
		// 
		if ( drawPos.x >= finishLine - [nekoImage size].width && ![neko finishedRace] )
		{
			drawPos.x = finishLine - [nekoImage size].width; 
			[neko setPosition: NSMakePoint(drawPos.x, 0)]; 
			[neko setFinishedRace]; 
			nekoImage = nekoImages[[neko imageToDisplay]];
			
			if ( !computerControlled )
			{
				gSoundPlayer.mPlaySound(kSoundCatFinished); 
			}
		}
		
		// draw the word for this neko (making sure if doesn't go outside of the view area)
		// 
		NSSize wordSize = [word size]; 
		float xAdjustment = 0; 
		if ( drawPos.x + wordSize.width > finishLine )
		{
			xAdjustment = (drawPos.x + wordSize.width) - finishLine; 
		}
		
		[word drawAtPoint: NSMakePoint(drawPos.x - xAdjustment, drawPos.y)]; 
		
		// adjust for the height of the text before drawing the image
		// 
		drawPos.y += fontHeight; 
		[nekoImage compositeToPoint: drawPos operation: NSCompositeSourceOver]; 
		
		// if this neko has a powerup in its path, draw it as well as checking to see if it has been
		// reached
		// 
		if ( powerupType < kPowerupMax && neko == nekoWithPowerup )
		{
			[powerupImages[powerupType] compositeToPoint: NSMakePoint(powerupPosition, drawPos.y) operation: NSCompositeSourceOver]; 
			
			if ( [neko position].x + [nekoImage size].width > powerupPosition )
			{
				switch( powerupType )
				{
				case kPowerupDistanceBoost:
				case kPowerupShield:
					[neko powerupEffect: powerupType withSound: !computerControlled];
					break; 
				
				case kPowerupForceSleep:
				case kPowerupGibberish:
					[parentView powerdownEffect: powerupType fromTeam: self]; 
					break; 
				
				default:
					break; 
				}
				
				powerupType = kPowerupMax; 
			}
		}
		
		// if a powerup is in effect for this neko, draw it as well
		// 
		if ( [neko currentPowerup] < kPowerupMax )
		{
			[powerupImages[[neko currentPowerup]] compositeToPoint: NSMakePoint(kCurrentPowerupDrawPosition, drawPos.y) 
													operation: NSCompositeSourceOver]; 
		}
		else if ( [neko shieldedFromBadPowerup] )
		{
			[powerupImages[kPowerupShield] compositeToPoint: NSMakePoint(kCurrentPowerupDrawPosition, drawPos.y) 
											operation: NSCompositeSourceOver]; 
		}
		 
		// set up to draw the next neko in the list
		// 
		yAdjustment += fontHeight + [nekoImage size].height; 
	}
	
	return yAdjustment; 
}

- (void)resetTeam: (BOOL)newGame isComputerControlled: (BOOL)computerControl
{
	computerControlled = computerControl; 
	
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		[neko reset]; 
	}
	
	if ( computerControlled )
	{
		selectedComputerNeko = nil; 
		charactersMatched = 0; 
		updateCharactersMatchedAtTime = 0;
		catnipped = NO; 
		
		if ( newGame )
		{
			switch( [[NSUserDefaults standardUserDefaults] integerForKey: NekoTypeRInitialDifficultyKey] )
			{
			case kDifficultyEasy:
				updateCharactersMatchedInterval = kBaseUpdateCharactersAtInterval + (kUpdateCharsDifficultyModifier * 6); 
				break;
			
			case kDifficultyNormal:
				updateCharactersMatchedInterval = kBaseUpdateCharactersAtInterval; 
				break;
			
			case kDifficultyHard:
				updateCharactersMatchedInterval = kBaseUpdateCharactersAtInterval - (kUpdateCharsDifficultyModifier * 3); 
				break;
			
			default:
				break;
			}
		}
	}
	
	if ( newGame )
	{
		powerupGenerationInterval = 1000 * (UInt32)[[NSUserDefaults standardUserDefaults] floatForKey: NekoTypeRBonusFrequencyKey]; 
		chanceOfPowerupAtInterval = [[NSUserDefaults standardUserDefaults] integerForKey: NekoTypeRBonusChanceKey]; 
		numberOfWordsTyped = 0; 
		leftoverTimeInMinute = kUpdateWPMTimeInterval; 
	}
	
	waitingForRaceToStart = YES; 
	powerupType = kPowerupMax; 
	updateWordsPerMinuteAtTime = 0; 
}

- (SInt32)matchCharactersTyped: (NSString *)charactersTyped
{
	SInt32 matchCode = kMatchNoMatchFound; 
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		if ( [neko isMoving] )
		{
			continue; 
		}
		
		NSString* word = [neko valueForKey: @"wordToType"]; 
		if ( [word hasPrefix: charactersTyped] )
		{
			[neko updateWordToDisplay: [charactersTyped length]];  
			matchCode = kMatchPartialMatchFound; 
			
			NSString* word = [neko valueForKey: @"wordToType"]; 
			if ( [word length] == [charactersTyped length] )
			{
				[neko moveNekoForward: -1]; 
				numberOfWordsTyped++; 
				matchCode = kMatchCompleteMatchFound; 
			}
		}
		else
		{
			[neko updateWordToDisplay: 0]; 
		}
	}
	
	return matchCode; 
}

- (BOOL)finishedRace
{
	BOOL allFinished = YES;
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		if ( ![neko finishedRace] )
		{
			allFinished = NO; 
		}
	}
	
	return allFinished; 
}

- (void)startComputerControl
{
	updateCharactersMatchedAtTime = gTimer.mGetMilliseconds() + updateCharactersMatchedInterval; 
}

- (void)updateComputerControl
{
	if ( selectedComputerNeko )
	{
		if ( updateCharactersMatchedAtTime < gTimer.mGetMilliseconds() )
		{
			[selectedComputerNeko updateWordToDisplay: ++charactersMatched]; 
			[parentView setNeedsDisplay: YES]; 
			
			updateCharactersMatchedAtTime = gTimer.mGetMilliseconds() + updateCharactersMatchedInterval; 
			if ( catnipped )
			{
				updateCharactersMatchedAtTime += updateCharactersMatchedInterval; 
			}
			
			NSString* wordToType = [selectedComputerNeko valueForKey: @"wordToType"]; 
			if ( [wordToType length] == charactersMatched )
			{
				[selectedComputerNeko moveNekoForward: -1];
				selectedComputerNeko = nil; 
				charactersMatched = 0; 
				numberOfWordsTyped++; 
				catnipped = NO; 
			}
		}
	}
	else
	{
		if ( [self finishedRace] )
		{
			[self raceEnded]; 
			selectedComputerNeko = nil; 
			charactersMatched = 0; 
			return;
		}
		
		selectedComputerNeko = [teamMembers objectAtIndex: uirand() % [teamMembers count]]; 
		if ( [selectedComputerNeko finishedRace] || [selectedComputerNeko isMoving] )
		{
			selectedComputerNeko = nil; 
		}
	}
}

- (void)increaseComputerDifficulty
{
	updateCharactersMatchedInterval -= kUpdateCharsDifficultyModifier; 
	if ( updateCharactersMatchedInterval < 0 )
	{
		updateCharactersMatchedInterval = 0; 
	}
}

- (void)startRacing
{
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		[neko moveNekoForward: -1]; 
	}
}

- (void)raceEnded
{
	if ( computerControlled )
	{
		computerControlled = NO; 
	}
	
	leftoverTimeInMinute = updateWordsPerMinuteAtTime - gTimer.mGetMilliseconds(); 
}

- (void)powerdownEffect: (SInt32)type withSound: (BOOL)sound
{
	if ( computerControlled )
	{
		switch( type )
		{
		case kPowerupForceSleep:
			selectedComputerNeko = nil; 
			charactersMatched = 0; 
			updateCharactersMatchedAtTime = 0;
			break;
		
		case kPowerupGibberish: 
			catnipped = YES; 
			selectedComputerNeko = nil;
			charactersMatched = 0; 
			updateCharactersMatchedAtTime = 0;
			break;
		
		default:
			break; 
		}
	}
	
	NSMutableArray* nekoList = [NSMutableArray arrayWithCapacity: 3]; 
	NekoTyper* neko = nil; 
	NSEnumerator* nekoIter = [teamMembers objectEnumerator]; 
	while ( neko = (NekoTyper *)[nekoIter nextObject] )
	{
		if ( ![neko finishedRace] )
		{
			[nekoList addObject: neko]; 
		}
	}
	
	if ( [nekoList count] )
	{
		NekoTyper* affectedNeko = [nekoList objectAtIndex: uirand() % [nekoList count]]; 
		[affectedNeko powerupEffect: type withSound: sound]; 
	}
}

@end
