//
//  NekoView.mm
//  Neko Type-R

/*
Copyright (c) 2005 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#import "NekoView.h"
#import "NekoController.h"
#import "NekoTyper.h"
#import "NekoTypingTeam.h"
#import "WordDispenser.h"
#include "Timer.h"
#include "randrot-b.h"
#include "SoundPlayer.h"
#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>

#define kPhaseOnMarkTimeInterval	1500
#define kPhaseGetSetTimeInterval	1500
#define kPhaseGoTimeInterval		750
#define kSpacingBetweenTeams		8.0f


@implementation NekoView

- (id)initWithFrame: (NSRect)frame 
{
	self = [super initWithFrame:frame];
	if (self) 
	{
		sdrand(time(NULL)); 
		gTimer.mInitializeTimer(); 
		
		controller = nil; 
		charactersTyped = [[NSMutableString alloc] initWithCapacity: 255]; 
		nekoUpdater = nil; 
		startPhase = kStartPhaseOnMark; 
		
		NSMutableArray* wordList = [self processTextFile: @"wordList"]; 
		words = [[WordDispenser alloc] init]; 
		[words setValue: wordList forKey: @"words"]; 
		
		catNames = [self processTextFile: @"catNames"]; 
		
		// create the two teams that will be racing against each other
		// 
		playerTeam = [[NekoTypingTeam alloc] init]; 
		[playerTeam loadTeamImages: @"Neko"]; 
		[playerTeam setValue: self forKey: @"parentView"]; 
		
		computerTeam = [[NekoTypingTeam alloc] init]; 
		[computerTeam loadTeamImages: @"Tabby"]; 
		[computerTeam setValue: self forKey: @"parentView"]; 
		
		// create the neko typers themselves
		// 
		SInt32 i; 
		for (i = 0; i < 6; i++)
		{
			NekoTyper* neko = [[NekoTyper alloc] init]; 
			[neko setValue: words forKey: @"words"]; 
			
			UInt32 catNameIdx = uirand() % [catNames count]; 
			[neko setValue: [catNames objectAtIndex: catNameIdx] forKey: @"name"]; 
			[catNames removeObjectAtIndex: catNameIdx]; 
			
			if ( i < 3 )
			{
				[playerTeam addTeamMember: neko]; 
			}
			else
			{
				[computerTeam addTeamMember: neko];
			}
		}
		
		startPhase = kStartPhaseMax; 
		nextStartPhaseAtTime = 0; 
		gameOver = YES; 
		
		introScreenImage = [NSImage imageNamed: @"intro_screen.png"]; 
		startPhaseImages[kStartPhaseOnMark] = [NSImage imageNamed: @"phase-on_mark.png"]; 
		startPhaseImages[kStartPhaseGetSet] = [NSImage imageNamed: @"phase-get_set.png"]; 
		startPhaseImages[kStartPhaseGo] = [NSImage imageNamed: @"phase-go.png"]; 
		
		gSoundPlayer.mPreloadSound(kSoundIntro, "intro.wav"); 
		gSoundPlayer.mPreloadSound(kSoundPlayerFinished, "player-finished.wav"); 
		gSoundPlayer.mPreloadSound(kSoundGameOver, "game-over.wav"); 
		gSoundPlayer.mPreloadSound(kSoundDistanceBoost, "powerup-dist_boost.wav"); 
		gSoundPlayer.mPreloadSound(kSoundForceSleep, "powerup-force_sleep.wav"); 
		gSoundPlayer.mPreloadSound(kSoundGibberish, "powerup-gibberish.wav"); 
		gSoundPlayer.mPreloadSound(kSoundShield, "powerup-shield.wav"); 
		gSoundPlayer.mPreloadSound(kSoundShieldUsed, "shield-used.wav"); 
		gSoundPlayer.mPreloadSound(kSoundCatFinished, "cat_finished.wav"); 
		gSoundPlayer.mPreloadSound(kSoundComputerPowerdown, "computer_powerdown.wav"); 
		gSoundPlayer.mPreloadSound(kSoundIncorrectLetter, "incorrect_letter.wav"); 
		gSoundPlayer.mPreloadSound(kSoundPhaseOnMark, "phase-on_mark.wav"); 
		gSoundPlayer.mPreloadSound(kSoundPhaseGetSet, "phase-get_set.wav"); 
		gSoundPlayer.mPreloadSound(kSoundPhaseGo, "phase-go.wav");
		
		gSoundPlayer.mPlaySound(kSoundIntro); 
	}
	
	return self;
}

- (void)dealloc
{
	[charactersTyped release]; 
	[words release]; 
	[catNames release]; 
	[playerTeam release]; 
	[computerTeam release]; 
	
	[super dealloc]; 
}

- (void)drawRect: (NSRect)rect 
{
	if ( gameOver )
	{
		[introScreenImage compositeToPoint: NSMakePoint(0, 0) operation: NSCompositeSourceOver]; 
		return; 
	}
	
	float yAdjustment = [playerTeam drawTeamAtHeight: 0] + kSpacingBetweenTeams; 
	yAdjustment = [computerTeam drawTeamAtHeight: yAdjustment]; 
	
	if ( startPhase < kStartPhaseMax )
	{
		NSRect bounds = [self bounds]; 
		NSSize imageSize = [startPhaseImages[startPhase] size]; 
		NSPoint p = NSMakePoint((bounds.size.width / 2) - (imageSize.width / 2), 
								(bounds.size.height / 2) - (imageSize.height / 2)); 
		
		[startPhaseImages[startPhase] compositeToPoint: p operation: NSCompositeSourceOver]; 
	}
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)keyDown: (NSEvent *)theEvent
{
	if ( gameOver )
	{
		return; 
	}
	
	// keep track of the characters typed in so far
	// 
	[charactersTyped appendString: [[theEvent charactersIgnoringModifiers] substringToIndex: 1]]; 
	
	// see if it matches (or partially matches) any of the neko in the player's team
	// 
	SInt32 matchCode = [playerTeam matchCharactersTyped: charactersTyped]; 
	switch( matchCode )
	{
	case kMatchNoMatchFound:
		gSoundPlayer.mPlaySound(kSoundIncorrectLetter); 
	
	case kMatchCompleteMatchFound:
		[charactersTyped setString: @""];
	
	case kMatchPartialMatchFound:
		[self setNeedsDisplay: YES]; 
		break;
	
	default:
		break;
	}
}

- (NSMutableArray *)processTextFile: (NSString *)textFile
{
	NSMutableArray* lines = [[NSMutableArray alloc] init]; 
	NSString* filePath = [[NSBundle mainBundle] pathForResource: textFile ofType: @"txt"]; 
	std::string line;  
	std::ifstream lineStream([filePath UTF8String]); 
	while ( !lineStream.eof() )
	{
		std::getline(lineStream, line); 
		
		if ( line.length() )
		{
			[lines addObject: [NSString stringWithUTF8String: line.c_str()]]; 
		}
	}
	
	lineStream.close(); 
	return lines; 
}

- (void)startNextRace: (BOOL)newGame
{
	gameOver = NO; 
	
	if ( newGame )
	{
		raceNumber = 1; 
		[words reset]; 
	}
	else
	{
		raceNumber++; 
		[self increaseRaceDifficulty]; 
	}
	
	[playerTeam resetTeam: newGame isComputerControlled: NO]; 
	[computerTeam resetTeam: newGame isComputerControlled: YES]; 
	
	// activate the timer to handle updating the neko positions
	// 
	[nekoUpdater invalidate]; 
	nekoUpdater = [NSTimer scheduledTimerWithTimeInterval: 0.001
							target: self
							selector: @selector(handleUpdateNekoTeams)
							userInfo: nil
							repeats: YES];
	
	startPhase = kStartPhaseOnMark; 
	nextStartPhaseAtTime = gTimer.mGetMilliseconds() + kPhaseOnMarkTimeInterval; 
	gSoundPlayer.mPlaySound(kSoundPhaseOnMark); 
	
	[controller updateRaceMessage: [NSString stringWithFormat: @"Starting race #%d...", raceNumber]]; 
}

- (void)handleUpdateNekoTeams
{
	gTimer.mUpdateTimer(); 
	
	[playerTeam updateTeam];
	[computerTeam updateTeam]; 
	
	if ( nextStartPhaseAtTime && nextStartPhaseAtTime < gTimer.mGetMilliseconds() )
	{
		[self updateStartPhaseDisplay]; 
	}
	
	[self setNeedsDisplay: YES];
	
	// check if the race is done
	//
	if ( [computerTeam finishedRace] )
	{
		// use the functional API to display an alert sheet to retain 10.2 compatibility
		//
		NSString* message = [NSString stringWithFormat: @"Doh! You lost the race."]; 
		NSBeginAlertSheet(message, 
							@"Okay", 
							nil, 
							nil, 
							[self window], 
							self, 
							@selector(computerWonRaceSheetDidEnd:returnCode:contextInfo:), 
							NULL, 
							nil, 
							@"", 
							nil); 
		
		[playerTeam raceEnded]; 
		[computerTeam raceEnded]; 
		[nekoUpdater invalidate]; 
		nekoUpdater = nil; 
		gSoundPlayer.mPlaySound(kSoundGameOver);
	}
	else if ( [playerTeam finishedRace] )
	{
		gSoundPlayer.mPlaySound(kSoundPlayerFinished); 
		[playerTeam raceEnded];
		[computerTeam raceEnded]; 
		[nekoUpdater invalidate]; 
		nekoUpdater = nil; 
		
		// use the functional API to display an alert sheet to retain 10.2 compatibility
		// 
		NSString* message = @"Woo-hoo! You won this race. Now on to the next one!"; 
		NSBeginAlertSheet(message, 
							@"Okay",   
							nil, 
							nil, 
							[self window], 
							self, 
							@selector(playerWonRaceSheetDidEnd:returnCode:contextInfo:), 
							NULL, 
							nil, 
							@"", 
							nil);
	}
}

- (void)computerWonRaceSheetDidEnd: (NSWindow *)sheet returnCode: (int)returnCode contextInfo: (void *)contextInfo
{
	if ( returnCode == NSAlertDefaultReturn )
	{
		gameOver = YES; 
		[self setNeedsDisplay: YES]; 
	}
}

- (void)playerWonRaceSheetDidEnd: (NSWindow *)sheet returnCode: (int)returnCode contextInfo: (void *)contextInfo
{
	if ( returnCode == NSAlertDefaultReturn )
	{
		[self startNextRace: NO]; 
	}
}

- (void)updateStartPhaseDisplay
{
	switch(++startPhase)
	{
	case kStartPhaseGetSet:
		nextStartPhaseAtTime = gTimer.mGetMilliseconds() + kPhaseGetSetTimeInterval;
		gSoundPlayer.mPlaySound(kSoundPhaseGetSet); 
		break; 
	
	case kStartPhaseGo:
		nextStartPhaseAtTime = gTimer.mGetMilliseconds() + kPhaseGoTimeInterval;
		gSoundPlayer.mPlaySound(kSoundPhaseGo); 
		break; 
	
	default:
		[playerTeam startRacing];
		[computerTeam startRacing];
		nextStartPhaseAtTime = 0; 
		break; 
	}
}

- (void)increaseRaceDifficulty
{
	[computerTeam increaseComputerDifficulty]; 
	[words increaseMaxLengthToDispense: 1];
}

- (void)updateWordsPerMinute: (UInt32)wpm ofTeam: (NekoTypingTeam *)team
{
	if ( team == playerTeam )
	{
		[controller updatePlayerWordsPerMinute: wpm]; 
	}
	else if ( team == computerTeam )
	{
		[controller updateComputerWordsPerMinute: wpm]; 
	}
}

- (void)powerdownEffect: (UInt32)type fromTeam: (NekoTypingTeam *)team
{
	if ( team == playerTeam )
	{
		gSoundPlayer.mPlaySound(kSoundComputerPowerdown); 
		[computerTeam powerdownEffect: type withSound: NO]; 
	}
	else if ( team == computerTeam )
	{
		[playerTeam powerdownEffect: type withSound: YES]; 
	}
}

@end
