/*
SoundPlayer.h

Copyright (C) 2004 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#ifndef SoundPlayer_h
#define SoundPlayer_h   1

#include <Carbon/Carbon.h>
#include <map>

class tSoundData; 

typedef std::map<SInt32, tSoundData*>	tSoundDataMap; 
typedef tSoundDataMap::iterator			tSoundDataIter; 

class tSoundPlayer
{
public:
	tSoundPlayer(); 
	~tSoundPlayer(); 
	
	/*! \brief  Initializes OpenAL. 
	*/
	void	mInit(int* argc, char** argv); 
	
	/*! \brief  Plays a sound. 
		\param  soundToken ID of the sound to play. 
		\param  loop Whether or not the sound to play should be looped.
	*/
	void	mPlaySound(SInt32 soundToken, bool loop = false); 
	
	/*! \brief  Stops a sound. 
		\param  soundToken ID of the sound to stop playing. 
	*/
	void	mStopSound(SInt32 soundToken); 
	
	/*! \brief  Checks if a sound is playing. 
		\param  soundToken ID of the sound to check if playing.
		\return True if the sound is playing, false otherwise. 
	*/
	bool	mIsPlaying(SInt32 soundToken); 
	
	void	mPreloadSound(SInt32 soundToken, char* filename); 
	
	void	mSetSoundPosition(SInt32 soundToken, float x, float y, float z); 
	void	mGetSoundPosition(SInt32 soundToken, float &x, float &y, float &z); 
	void	mSetListenerPosition(float x, float y, float z); 
	void	mGetListenerPosition(float &x, float &y, float &z); 
	
protected:

	tSoundDataMap  fSounds; //!< Map structure of sound data to use when looking up the sound to play. 
	
	/*! \brief  Creates a sound data object. 
		\param  soundToken ID for the sound data to create. 
		\return A tSoundData object that contains the info needed to play the sound. 
	*/
	tSoundData* mCreateSoundDataObject(char* filename); 
};

extern tSoundPlayer gSoundPlayer; 

#endif
