/*
 *  Timer.h
 */

/*
Copyright (c) 2005 Vince Tagle <vtagle@newsguy.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of 
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
*/

#ifndef Timer_h
#define Timer_h 1

#include <sys/time.h>
#include "BasicTypes.h"

static struct timeval sTime; 

static void gInitializeTimer()
{
	gettimeofday(&sTime, 0); 
}

static inline unsigned long gGetMilliseconds()
{
	struct timeval	cTime; 
	unsigned long	result; 
	
	gettimeofday(&cTime, 0); 
	result = ((cTime.tv_sec - sTime.tv_sec) * 1000) + ((cTime.tv_usec - sTime.tv_usec) / 1000); 
	
	return result; 
}

static inline unsigned long gGetMicrosecondsOnly()
{
	struct timeval	cTime; 
	gettimeofday(&cTime, 0); 
	
	return cTime.tv_usec; 
}

class tTimer
{
public:

	tTimer(); 
	~tTimer(); 
	
	void	mInitializeTimer(); 
	UInt32	mGetMilliseconds(); 
	
	inline void		mUpdateTimer(); 
	inline UInt32	mGetMillisecondsSinceLastUpdate(); 
	inline double	mGetSecondsSinceLastUpdate(); 
	
protected:

	UInt32			fTimeOfLastUpdate; 
	UInt32			fMillisecondsSinceLastUpdate; 
	struct timeval	fTimeInit; 
};

inline void tTimer::mUpdateTimer()
{
	if ( !fTimeOfLastUpdate )
	{
		fTimeOfLastUpdate = mGetMilliseconds();
	}
	
	fMillisecondsSinceLastUpdate = mGetMilliseconds() - fTimeOfLastUpdate; 
	fTimeOfLastUpdate = mGetMilliseconds();
}

inline UInt32 tTimer::mGetMillisecondsSinceLastUpdate()
{
	return fMillisecondsSinceLastUpdate; 
}

inline double tTimer::mGetSecondsSinceLastUpdate()
{
	return fMillisecondsSinceLastUpdate / 1000.0f; 
}

extern tTimer gTimer; 

#endif
